"""
Act Like A Customer Tests.

These tests are specifically designed to test lots of the system in
not many tests.

Pros:
- Can test lots of the system quickly.

Cons:
- Bail at first failure without testing all actions
- Don't follow arrange-act-assert or given-when-then style and
  are therefore terrible examples of how to write a test
"""

import pytest
from cards import Card


@pytest.mark.alac
def test_alac_1(db_empty):
    """
    Starting from an empty list.
    1. Add 3 items
    2. Make sure they show up in the list.
    3. Change the owner on a card. Verify change.
    4. Change the done state on a card. Verify change.
    5. Change the sumary on a card. Verify change.
    6. Delete the done item. Verify change.
    """

    # Starting from an empty list.
    db = db_empty

    # 1. Add 3 items
    input = ({'summary': 'something'},
             {'summary': 'something else', 'owner': 'okken'},
             {'summary': 'Foo Bar Baz'})
    ids = []
    for i in input:
        ids.append(db.add(Card(**i)))  # ** is dictionary unpacking

    # 2. Make sure they show up in the list.
    all_cards = db.list_cards()

    # extra checks, not specifically called out in description but logical
    assert len(input) == db.count()
    assert len(input) == len(all_cards)

    # now compare list with input
    expected_cards = tuple(Card(**i) for i in input)
    for expected, card in zip(expected_cards, all_cards):  # pairwise
        assert expected == card
    # assert expected_cards == all_cards # this doesn't work, not sure why

    # 3. Change the owner on a card. Verify change.
    a_valid_id = ids[0]
    card_before = db.get(a_valid_id)
    db.update(a_valid_id, Card(owner='okken'))
    card_after = db.get(a_valid_id)

    assert card_before != card_after  # make sure a change happened
    assert card_after.owner == 'okken'

    # 4. Change the done state on a card. Verify change.
    a_valid_id = ids[1]
    card_before = db.get(a_valid_id)
    db.update(a_valid_id, Card(done=True))
    card_after = db.get(a_valid_id)

    assert card_before != card_after  # make sure a change happened
    assert card_after.done

    # 5. Change the sumary on a card. Verify change.
    a_valid_id = ids[2]
    card_before = db.get(a_valid_id)
    db.update(a_valid_id, Card(summary='just sit'))
    card_after = db.get(a_valid_id)

    assert card_before != card_after  # make sure a change happened
    assert card_after.summary == 'just sit'

    # 6. Delete the done item. Verify change.

    # delete all done items, there should be just one
    # maybe this should be a utility API function?
    for c in db.list_cards():
        if c.done:
            db.delete(c.id)

    expected_cards = (Card(summary='something', owner='okken'),
                      Card(summary='just sit'))
    all_cards = db.list_cards()
    for expected, card in zip(expected_cards, all_cards):
        assert expected == card



from click.testing import CliRunner
import pytest
import pathlib
import cards.cli


@pytest.fixture()
def db_empty(tmpdir, monkeypatch):
    fake_home = pathlib.Path(str(tmpdir.mkdir('fake_home')))

    class FakePathLibPath():
        def home(self):
            return fake_home

    monkeypatch.setattr(cards.cli.pathlib, 'Path', FakePathLibPath)


@pytest.fixture()
def runner():
    return CliRunner()


@pytest.fixture()
def db_non_empty(db_empty, runner):
    runner.invoke(cards.cli.cards_cli, ['add', 'first item'])
    runner.invoke(cards.cli.cards_cli, ['add', 'second item'])
    runner.invoke(cards.cli.cards_cli, ['add', 'third item'])



import pytest
from cards import CardsDB, Card


@pytest.fixture(scope='session')
def db_session(tmpdir_factory):
    db_path = tmpdir_factory.mktemp('data').join('.cards_db.json')
    db = CardsDB(db_path)
    assert 0 == db.count()
    return db


@pytest.fixture()
def db_empty(db_session):
    db = db_session
    db.delete_all()
    assert 0 == db.count()
    return db


@pytest.fixture()
def db_non_empty(db_session):
    db = db_session
    db.add(Card(summary='first item'))
    db.add(Card(summary='second item'))
    db.add(Card(summary='third item'))
    return db



"""
Act Like A Customer Tests.

These tests are specifically designed to test lots of the system in
not many tests.

Pros:
- Can test lots of the system quickly.

Cons:
- Bail at first failure without testing all actions
- Don't follow arrange-act-assert or given-when-then style and
  are therefore terrible examples of how to write a test
"""

import pytest
from cards.cli import cards_cli


pytestmark = pytest.mark.cli


@pytest.mark.alac
def test_alac_1(db_empty, runner):
    """
    Starting from an empty list.
    1. Add 3 items
    2. Make sure they show up in the list.
    3. Change the owner on a card. Verify change.
    4. Change the done state on a card. Verify change.
    5. Change the sumary on a card. Verify change.
    6. Delete the done item. Verify change.
    """

    # Starting from an empty list.

    # 1. Add 3 items
    runner.invoke(cards_cli, ['add', 'something'])
    runner.invoke(cards_cli, ['add', 'something else', '-o', 'okken'])
    runner.invoke(cards_cli, ['add', 'Foo Bar Baz'])

    # 2. Make sure they show up in the list.
    result = runner.invoke(cards_cli, ['list'])
    expected = ("  ID      owner  done summary\n"
                "  --      -----  ---- -------\n"
                "   1                  something\n"
                "   2      okken       something else\n"
                "   3                  Foo Bar Baz\n")
    assert expected == result.output

    # 3. Change the owner on a card. Verify change.
    runner.invoke(cards_cli, ['update', '1', '-o', 'okken'])
    result = runner.invoke(cards_cli, ['list'])
    expected = ("  ID      owner  done summary\n"
                "  --      -----  ---- -------\n"
                "   1      okken       something\n"
                "   2      okken       something else\n"
                "   3                  Foo Bar Baz\n")
    assert expected == result.output

    # 4. Change the done state on a card. Verify change.
    runner.invoke(cards_cli, ['update', '2', '-d', 'True'])
    result = runner.invoke(cards_cli, ['list'])
    expected = ("  ID      owner  done summary\n"
                "  --      -----  ---- -------\n"
                "   1      okken       something\n"
                "   2      okken    x  something else\n"
                "   3                  Foo Bar Baz\n")
    assert expected == result.output

    # 5. Change the sumary on a card. Verify change.
    runner.invoke(cards_cli, ['update', '3', '-s', 'just sit'])
    result = runner.invoke(cards_cli, ['list'])
    expected = ("  ID      owner  done summary\n"
                "  --      -----  ---- -------\n"
                "   1      okken       something\n"
                "   2      okken    x  something else\n"
                "   3                  just sit\n")
    assert expected == result.output

    # 6. Delete the done item. Verify change.
    runner.invoke(cards_cli, ['delete', '2'])
    result = runner.invoke(cards_cli, ['list'])
    expected = ("  ID      owner  done summary\n"
                "  --      -----  ---- -------\n"
                "   1      okken       something\n"
                "   3                  just sit\n")
    assert expected == result.output

    # let's also check the count
    result = runner.invoke(cards_cli, ['count'])
    expected = "2\n"
    assert expected == result.output



# -*- coding: utf-8 -*-
"""
Tests using cards CLI (command line interface).
This file contains tracer bullet tests.
Not in depth tests of functionality.
"""

import pytest
import cards.cli


pytestmark = pytest.mark.cli


@pytest.mark.smoke
def test_add(db_empty, runner):
    # GIVEN an empty database
    # WHEN a new card is added
    runner.invoke(cards.cli.cards_cli, ['add', 'something'])

    # THEN The listing returns just the new card
    result = runner.invoke(cards.cli.cards_cli, ['list'])
    expected_output = ("  ID      owner  done summary\n"
                       "  --      -----  ---- -------\n"
                       "   1                  something\n")
    assert expected_output == result.output


@pytest.mark.smoke
def test_list(db_empty, runner):
    # GIVEN a db with known contents of 2 cards
    runner.invoke(cards.cli.cards_cli, ['add', 'one'])
    runner.invoke(cards.cli.cards_cli, ['add', 'two'])

    # `cards list` returns our 2 cards
    result = runner.invoke(cards.cli.cards_cli, ['list'])
    expected = ("  ID      owner  done summary\n"
                "  --      -----  ---- -------\n"
                "   1                  one\n"
                "   2                  two\n")
    assert expected == result.output


@pytest.mark.smoke
def test_count(db_empty, runner):
    # GIVEN a db with 2 cards
    runner.invoke(cards.cli.cards_cli, ['add', 'one'])
    runner.invoke(cards.cli.cards_cli, ['add', 'two'])

    # `cards count` returns 2
    result = runner.invoke(cards.cli.cards_cli, ['count'])
    expected = "2\n"
    assert expected == result.output


# this test was annoying to write. parsing list output is yucky
@pytest.mark.smoke
def test_update(db_non_empty, runner):
    # GIVEN a card known to be in the db
    result = runner.invoke(cards.cli.cards_cli, ['list'])

    # this is kinda tricky
    last_item_as_list = result.output.split('\n')[-2].split()
    orig_id = last_item_as_list[0]
    orig_summary = ' '.join(last_item_as_list[1:])

    # WHEN we `cards update` the card with new info
    runner.invoke(cards.cli.cards_cli,
                  ['update', '-o', 'okken', '-d', 'True', orig_id])

    # THEN `cards list` will show the changes
    result = runner.invoke(cards.cli.cards_cli, ['list'])
    last_item_as_list = result.output.split('\n')[-2].split()
    id = last_item_as_list[0]
    owner = last_item_as_list[1]
    done = last_item_as_list[2]
    summary = ' '.join(last_item_as_list[3:])
    assert orig_id == id
    assert owner == 'okken'
    assert done == 'x'
    assert orig_summary == summary


@pytest.mark.smoke
def test_delete(db_empty, runner):
    # GIVEN a db with 2 items
    runner.invoke(cards.cli.cards_cli, ['add', 'one'])
    runner.invoke(cards.cli.cards_cli, ['add', 'two'])

    # WHEN we delete one item
    # hoping ids always start with 1
    runner.invoke(cards.cli.cards_cli, ['delete', '1'])

    # THEN the other card remains in the db
    result = runner.invoke(cards.cli.cards_cli, ['list'])
    expected = ("  ID      owner  done summary\n"
                "  --      -----  ---- -------\n"
                "   2                  two\n")
    assert expected == result.output



# -*- coding: utf-8 -*-
"""
Tests using cards CLI (command line interface).
This file contains tests related to design changes.
"""

import pytest
import cards.cli


pytestmark = pytest.mark.cli


def test_list_on_no_command(db_non_empty, runner):
    """Does `cards` behave just like `cards list`?"""
    output_with_list = runner.invoke(cards.cli.cards_cli, ['list']).output
    output_without_list = runner.invoke(cards.cli.cards_cli).output

    expected_output = ("  ID      owner  done summary\n"
                       "  --      -----  ---- -------\n"
                       "   1                  first item\n"
                       "   2                  second item\n"
                       "   3                  third item\n")
    assert expected_output == output_with_list
    assert expected_output == output_without_list
    assert output_with_list == output_without_list



"""Command Line Interface (CLI) for cards project."""

import click
import cards
import pathlib


@click.group(invoke_without_command=True,
             context_settings={'help_option_names': ['-h', '--help']})
@click.version_option(version=cards.__version__)
@click.pass_context
def cards_cli(ctx):
    """Run the cards application."""
    if ctx.invoked_subcommand is None:
        ctx.invoke(list_cards)


@cards_cli.command(help="add a card")
@click.argument('summary')
@click.option('-o', '--owner', default=None,
              help='set the card owner')
def add(summary, owner):
    """Add a card to db."""
    cards_db().add(cards.Card(summary, owner))


@cards_cli.command(help="delete a card")
@click.argument('card_id', type=int)
def delete(card_id):
    """Remove card in db with given id."""
    cards_db().delete(card_id)


@cards_cli.command(name="list", help="list cards")
def list_cards():
    """
    List cards in db.

    If owner given, only list cards with that owner.
    """
    formatstr = "{: >4} {: >10} {: >5} {}"
    print(formatstr.format('ID', 'owner', 'done', 'summary'))
    print(formatstr.format('--', '-----', '----', '-------'))
    for t in cards_db().list_cards():
        done = ' x ' if t.done else ''
        owner = '' if t.owner is None else t.owner
        print(formatstr.format(
              t.id, owner, done, t.summary))


@cards_cli.command(help="update card")
@click.argument('card_id', type=int)
@click.option('-o', '--owner', default=None,
              help='change the card owner')
@click.option('-s', '--summary', default=None,
              help='change the card summary')
@click.option('-d', '--done', default=None,
              type=bool,
              help='change the card done state (True or False)')
def update(card_id, owner, summary, done):
    """Modify a card in db with given id with new info."""
    cards_db().update(card_id, cards.Card(summary, owner, done))


@cards_cli.command(help="list count")
def count():
    """Return number of cards in db."""
    print(cards_db().count())


def cards_db():
    # just put it in users home dir for now
    db_path = pathlib.Path().home() / '.cards_db.json'
    return cards.CardsDB(db_path)



# -*- coding: utf-8 -*-
"""
Tests using cards API.
This file contains tracer bullet tests.
Not in depth tests of functionality.
"""

import pytest
from cards import Card


@pytest.mark.smoke
def test_add(db_empty):
    # GIVEN an empty database
    db = db_empty

    # WHEN a new card is added
    new_card = Card('do something')
    db.add(new_card)

    # THEN The listing returns just the new card
    all_cards = db.list_cards()
    assert 1 == len(all_cards)
    assert new_card == all_cards[0]


@pytest.mark.smoke
def test_get(db_non_empty):
    # GIVEN a card we know is in the db
    db = db_non_empty
    all_cards = db.list_cards()
    a_card = all_cards[0]

    # WHEN we call get() with the id of the known card
    retrieved_card = db.get(a_card.id)

    # THEN the retrieved card is identical
    # AND has the same id
    assert a_card == retrieved_card
    assert a_card.id == retrieved_card.id


@pytest.mark.smoke
def test_list(db_empty):
    # GIVEN a db with known contents of 2 cards
    db = db_empty
    db.add(Card(summary='one'))
    db.add(Card(summary='two'))

    # list_cards() returns our 2 cards
    all_cards = db.list_cards()
    expected = [Card(summary='one'), Card(summary='two')]
    assert expected == all_cards


@pytest.mark.smoke
def test_count(db_empty):
    # GIVEN a db with 2 cards
    db = db_empty
    db.add(Card(summary='one'))
    db.add(Card(summary='two'))

    # count() returns 2
    count = db.count()
    assert 2 == count


@pytest.mark.smoke
def test_update(db_non_empty):
    # GIVEN a card known to be in the db
    db = db_non_empty
    all_cards = db.list_cards()
    a_card = all_cards[0]

    # WHEN we update() the card with new info
    db.update(a_card.id, Card(owner='okken', done=True))

    # THEN we can retrieve the card with get() and
    # and it has all of our changes
    updated_card = db.get(a_card.id)
    expected = Card(summary=a_card.summary, owner='okken', done=True)
    assert expected == updated_card


@pytest.mark.smoke
def test_delete(db_empty):
    # GIVEN a db with 2 items
    db = db_empty
    id_1 = db.add(Card(summary='one'))
    db.add(Card(summary='two'))

    # WHEN we delete one item
    db.delete(id_1)

    # THEN the other card remains in the db
    all_cards = db.list_cards()
    expected = [Card(summary='two')]
    assert expected == all_cards


@pytest.mark.smoke
def test_delete_all(db_non_empty):
    # GIVEN a non empty db
    db = db_non_empty

    # WHEN we delete_all()
    db.delete_all()

    # THEN the count is 0
    count = db.count()
    assert 0 == count



# -*- coding: utf-8 -*-

"""
cardsdb :  The API
"""

import attr
import tinydb
from typing import List


@attr.s
class Card(object):
    summary: str = attr.ib(default=None)
    owner: str = attr.ib(default=None)
    done: bool = attr.ib(default=None)
    id: int = attr.ib(default=None, cmp=False)

    @classmethod
    def from_dict(cls, d):
        return Card(**d)  # dictionary unpacking as of Python 3.5, PEP 448

    def to_dict(self):
        return attr.asdict(self)

# --- actions on db


class CardsDB():

    def __init__(self, db_path):
        self._db = tinydb.TinyDB(db_path)

    def add(self, card: Card) -> int:
        """Add a card to the db."""
        card.id = self._db.insert(card.to_dict())
        self._db.update(card.to_dict(), doc_ids=[card.id])
        return card.id

    def get(self, card_id: int) -> Card:
        """Return a card with a matching id."""
        return Card.from_dict(self._db.get(doc_id=card_id))

    def list_cards(self) -> List[Card]:
        """Return a list of all cards."""
        return [Card.from_dict(t) for t in self._db]

    def count(self) -> int:
        """Return the number of cards in db."""
        return len(self._db)

    def update(self, card_id: int, card_mods: Card) -> None:
        """Update a card with modifications."""
        d = card_mods.to_dict()
        changes = {k: v for k, v in d.items() if v is not None}
        self._db.update(changes, doc_ids=[card_id])

    def delete(self, card_id: int) -> None:
        """Remove a card from db with given card_id."""
        self._db.remove(doc_ids=[card_id])

    def delete_all(self) -> None:
        """Remove all tasks from db."""
        self._db.purge()



# !/usr/bin/env python
# -*- coding: utf-8 -*-

import io
import os
import re

from setuptools import find_packages, setup, Command

# Package meta-data.
NAME = 'cards'
DESCRIPTION = 'Project task tracking / todo list.'
URL = 'https://github.com/okken/cards'
EMAIL = 'brian@pythontesting.net'
AUTHOR = 'Brian Okken'
REQUIRES_PYTHON = '>=3.6.0'
VERSION = '0.1.4'

REQUIRED = [ 'Click==6.7', 'tinydb==3.8.0', 'attrs==17.4.0' ]

# The rest you shouldn't have to touch too much :)
# ------------------------------------------------
# Except, perhaps the License and Trove Classifiers!
# If you do change the License, remember to change the Trove Classifier for that!

here = os.path.abspath(os.path.dirname(__file__))

# Import the README and use it as the long-description.
with io.open(os.path.join(here, 'README.rst'), encoding='utf-8') as f:
    long_description = '\n' + f.read()

def read(*parts):
    with io.open(os.path.join(here, *parts), 'r') as fp:
        return fp.read()

def find_version(*file_paths):
    version_file = read(*file_paths)
    version_match = re.search(r"^__version__ = ['\"]([^'\"]*)['\"]",
                              version_file, re.M)
    if version_match:
        return version_match.group(1)
    raise RuntimeError("Unable to find version string.")


setup(
    name=NAME,
    version=find_version('cards', '__init__.py'),
    description=DESCRIPTION,
    long_description=long_description,
    author=AUTHOR,
    author_email=EMAIL,
    python_requires=REQUIRES_PYTHON,
    url=URL,
    packages=find_packages(exclude=('tests',)),
    entry_points={
        'console_scripts': ['cards=cards.cli:cards_cli'],
    },
    install_requires=REQUIRED,
    include_package_data=True,
    license='MIT',
    classifiers=[
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: Implementation :: CPython',
        'Programming Language :: Python :: Implementation :: PyPy'
    ],
)
